local MINOR_VERSION = tonumber(("$Revision: 39531 $"):match("%d+"))

Omen =  AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
Omen:RegisterDB("OmenDB", "OmenDBPC")
Omen:RegisterDefaults('profile', {
	ShowOmen = true,
	AnimateBars = true,
	StretchBarTextures = false,
	BarTexture = "Glaze",
	TPSUpdateFrequency = 0.05,
	TPSPrecision = "%2.1f",
	TPSWindow = 10,
	GrowUp = false,
	MaxBars = 10,
	BarHeight = 15,
	AlwaysShowSelf = true,
	ActivateSolo = false,
	ShowKtmData = true,
	ShowTPS = true,
	Width = 210,
	BackdropColor = {r = 0, g = 0, b = 0, a = 0.5},
	BorderColor = {r = 0, g = 0, b = 0, a = 0.5},
	PlayerBarColor = {r = 1, g = 0, b = 0, a = 1},
	UseCustomBarColors = false,
	CustomColors = {},
	CustomBarTextures = {},
	CustomBarHeights = {},
	ThreatBars = {},
	Scale = 1,
	ColumnVisibility = {
		name = true,
		tps = true,
		threat = true,
		threatpercent = true
	},
	ColumnOffsets = {
		name = {0.03, "LEFT", "LEFT"},
		tps = {0.5, "LEFT", "RIGHT"},
		threat = {0.33, "LEFT", "RIGHT"},
		threatpercent = {0.03, "RIGHT", "RIGHT"},
	}
})

local L = AceLibrary("AceLocale-2.2"):new("Omen")
local dewdrop = AceLibrary("Dewdrop-2.0")

local PlayerName = UnitName("player")

local Threat = AceLibrary("Threat-1.0")
local SML = AceLibrary("SharedMedia-1.0")

Omen.version = "2.1r" .. MINOR_VERSION
Omen.revision = MINOR_VERSION

local new, newHash, newSet, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	function newSet(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	function del(t)
		setmetatable(t, nil)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local tpsSigma = {}
local tpsVals = {}
local lastThreat = {}
local threatBars = {}

function Omen:OnInitialize()
	self.bars = {}
	
	self:RegisterEvent("Threat_ThreatCleared", "Clear")
	self:RegisterEvent("Threat_PartyChanged")
	self:RegisterEvent("Threat_OutOfDateNotice")
	self:RegisterEvent("AceEvent_FullyInitialized")

	self.hasIcon = true

	self.Anchor = CreateFrame("Frame", "OmenAnchor", UIParent)
	
	self.Anchor:SetResizable(true)
	local grip = CreateFrame("Button", "OmenResizeGrip", self.Anchor)
	grip:SetNormalTexture("Interface\\AddOns\\Omen\\ResizeGrip")
	grip:SetHighlightTexture("Interface\\AddOns\\Omen\\ResizeGrip")
	grip:SetWidth(16)
	grip:SetHeight(16)
	grip:SetScript("OnMouseDown", function()
		if not Omen.db.profile.Locked then
			this:GetParent():StartSizing()
		end
	end)
	grip:SetScript("OnMouseUp", function()
		this:GetParent():StopMovingOrSizing()
	end)
	grip:SetPoint("BOTTOMRIGHT", self.Anchor, "BOTTOMRIGHT", 0, 1)
	self.Grip = grip
	
	self.Anchor:SetWidth(self.db.profile.Width)
	if self.db.profile.GrowUp then
		self.Anchor:SetPoint("BOTTOM", UIParent, "CENTER", 0, 0)
	else
		self.Anchor:SetPoint("TOP", UIParent, "CENTER", 0, 0)
	end
	self.Anchor:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = true,
		tileSize = 16,
		edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4}
	})
	self.Anchor:SetClampedToScreen(true)
	self.Anchor.title = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	-- self.Anchor.title:SetPoint("TOPLEFT", self.Anchor, "TOPLEFT", 7, -7)

	-- self.Anchor.mobName = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	self.Anchor.title:SetPoint("BOTTOM", self.Anchor, "TOP", 0, 2)

	self.Anchor.revision = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	self.Anchor.revision:SetPoint("TOPLEFT", self.Anchor, "BOTTOMLEFT", 4, 2)
	local font, size = self.Anchor.revision:GetFont()
	self.Anchor.revision:SetFont(font, 8)
	self.modName = "|cff6299FFOmen"
	
	self.Anchor.revision:SetText(string.format("|cffFFB400Omen r |cffffffff%s /|cffFFB400 Threat-1.0 r |cffffffff%s", MINOR_VERSION, select(2,Threat:GetLibraryVersion())))

	self.Anchor.outofDate = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	self.Anchor.outofDate:SetPoint("TOPLEFT", self.Anchor.revision, "BOTTOMLEFT", 0, -1)
	self.Anchor.outofDate:Hide()
	local font, size = self.Anchor.revision:GetFont()
	self.Anchor.outofDate:SetFont(font, 8)
	self.Anchor.outofDate:SetText("|cffff0000Out of date notice")
	
	self.threatTargetName = ""
	self.Anchor.title:SetText(self.modName)
	local f = self.db.profile.BackdropColor
	self.Anchor:SetBackdropColor(f.r, f.g, f.b, f.a)

	local f = self.db.profile.BorderColor
	self.Anchor:SetBackdropBorderColor(f.r, f.g, f.b, f.a)

	self.titleOffset = 0
	self.barsOffset = 22 + self.titleOffset

	self.Anchor:SetMinResize(190, self.barsOffset + 6)
	self.Anchor:SetScript("OnSizeChanged", function()
		Omen:UpdateLayout()
	end)
	
	--[[ 
	self.LockButton = CreateFrame("Button", "OmenLockButton", self.Anchor)
	self.LockButton:SetNormalTexture("Interface/Icons/INV_Misc_Gear_06")
	self.LockButton:SetPushedTexture("Interface/Icons/INV_Misc_Gear_07")
	self.LockButton:SetPoint("TOPRIGHT", self.Anchor, "TOPRIGHT", -7, -7)
	self.LockButton:SetWidth(18)
	self.LockButton:SetHeight(18)
	self.LockButton:SetScript("OnClick", function()
		Omen.db.profile.Locked = not Omen.db.profile.Locked
		if Omen.db.profile.Locked then
			this:SetButtonState("PUSHED", true)
		else
			this:SetButtonState("NORMAL")
		end
	end)
	if Omen.db.profile.Locked then
		self.LockButton:SetButtonState("PUSHED", true)
	else
		self.LockButton:SetButtonState("NORMAL")
	end
	
	self.LockButton:SetScript("OnEnter", function()
		GameTooltip:SetOwner(self.LockButton, "ANCHOR_CENTER")
		GameTooltip:SetText(L["Pin/Unpin the Omen window"])
		GameTooltip:Show()
	end)
	self.LockButton:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	]]--
		
	-- INV_Gizmo_02.png -- config
	-- INV_Misc_Gear_06.png -- unpinned
	-- INV_Misc_Gear_07.png -- pinned
	-- INV_Misc_Note_04.png -- paper for chart

	self.Anchor:SetMovable(true)
	self.Anchor:EnableMouse(true)
	self.Anchor:SetScript("OnMouseDown", function() if not Omen.db.profile.Locked then this:StartMoving(); end end)
	self.Anchor:SetScript("OnMouseUp", function()
		this:StopMovingOrSizing();		
		Omen:SetAnchors()
	end)
	self.barList = {}
	self:Toggle(self.db.profile.ShowOmen)
	self:RegisterBucketEvent("Omen_ThreatUpdated", 0.2, "ArrangeBars")
	self:ArrangeBars()
	
	self.Anchor:SetHeight(self.barsOffset + 5)
	self.Anchor:SetScale(min(5, max(0.2, self.db.profile.Scale)))
	
	dewdrop:Register(self.Anchor, "children", self.OnMenuRequest)
	
	self.Anchor.columns = {}
	local barWidth = self.Anchor:GetWidth() - 10
	self:CreateColumn("name", L["Name"], 11, 10)
	self:CreateColumn("tps", L["TPS"], 11, 10)
	self:CreateColumn("threat", L["Threat"], 11, 10)
	self:CreateColumn("threatpercent", "%", 11, 10)
	for k in pairs(self.Anchor.columns) do
		self:ToggleColumn(k, self.db.profile.ColumnVisibility[k])
	end
	
	if self.db.profile.ActivateSolo then
		Threat:RequestActiveOnSolo(true)
	end
	
	if Threat:IsActive() then
		self:Threat_Activate()
	end

	self:RegisterEvent("Threat_Activate")
	self:RegisterEvent("Threat_Deactivate")		
	
	local columns = {L["Name"], "name", L["TPS"], "tps", L["Threat"], "threat", "%", "threatpercent"}
	for i = 1, #columns, 2 do
		local name = columns[i]
		local iName = columns[i+1]
		local val = {
			type = "group",
			name = name,
			desc = name,
			args = {
				["toggle_column_" .. iName] = {
					type = "toggle",
					name = L["Show"],
					desc = L["Show"],
					get = function()
						return Omen.db.profile.ColumnVisibility[iName]
					end,
					set = function(v)
						Omen.db.profile.ColumnVisibility[iName] = v
						Omen:ToggleColumn(iName, v)
					end
				},
				["column_offset_" .. iName] = {
					type = "range",
					name = L["Offset"],
					desc = L["Offset"],
					min = 0,
					max = 1,
					step = 0.01,
					isPercent = true,
					get = function()
						return Omen.db.profile.ColumnOffsets[iName][1]
					end,
					set = function(v)
						Omen.db.profile.ColumnOffsets[iName][1] = v
						Omen:UpdateLayout()
					end
				}
				
			}			
		}
		self.mainMenu.args.columns.args["column_" .. iName] = val
	end
	
	for k,v in pairs(self.db.profile.ThreatBars) do
		local bar = OmenThreatBar:new()
		bar:LoadBar(k)
		tinsert(threatBars, bar)
	end
	self:SetAnchors()
end

function Omen:AceEvent_FullyInitialized()
	self:CreateAggroGainBar("_AGGRO_GAIN_", string.format("*%s*", L["Aggro Gain"]))
	self:RegisterEvent("Threat_ThreatUpdated")
	for k,v in ipairs(threatBars) do
		v:UpdateTexture()
	end
end

local function ConvertToK(n)
	if n > 1000 then
		return n / 1000, "k"
	end
	return n, ""
end

function addBar(self, name, unit)
	if not self.bars[name] then
		self.bars[name] = self:CreateBar()
		self.bars[name].unitid = unit
		self.bars[name].unitname = name
		self:UpdateBarDisplay(name)
		tinsert(self.barList, self.bars[name])
	end
end

local function resetTPSSigmas(self)
	self:DisableTPSUpdate()
	for k in pairs(tpsSigma) do
		tpsSigma[k] = nil
	end
	for k in pairs(lastThreat) do
		lastThreat[k] = nil
	end
end

function Omen:ResetTPS()
	resetTPSSigmas(self)
end

function Omen:OnClick()
	if Threat:IsActive() then
		Omen:Toggle()
	end
end

function Omen:Threat_Activate()
	if not self:IsEventRegistered("UNIT_TARGET") then
		self:RegisterEvent("UNIT_TARGET")
	end
	if not self:IsEventRegistered("PLAYER_TARGET_CHANGED") then
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
	end
	
	self:SetIcon("Interface\\AddOns\\Omen\\icon")
	Omen:Toggle(self.db.profile.ShowOmen)
end

function Omen:Threat_Deactivate()
	if self:IsEventRegistered("UNIT_TARGET") then
		self:UnregisterEvent("UNIT_TARGET")
	end
	if self:IsEventRegistered("PLAYER_TARGET_CHANGED") then
		self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	end
	self:SetIcon("Interface\\AddOns\\Omen\\icon_off")
	self.Anchor:Hide()
end

function Omen:Toggle(state)
	if state or not self.Anchor:IsVisible() then
		self.Anchor:Show()
		self:ArrangeBars()
	else
		self.Anchor:Hide()
	end

	self.db.profile.ShowOmen = (Omen.Anchor:IsVisible() and true or false) -- DB didn't like nil
end

-- If our party changes around, we want to follow our target by name
-- We use the Threat event rather than the Blizzard event because this way we
-- know that Threat:GetUnitIDByUnitName will be accurate
function Omen:Threat_PartyChanged()
	if self.partyMemberName then
		local partyID = Threat:GetUnitIDByUnitName(self.partyMemberName)
		if partyID then
			self.watchedUnitID = partyID
			self.threatTargetName = UnitName(partyID .. "target")
		else
			-- We couldn't find a corresponding unit for the party member we want to watch
			-- Go ahead and invalidate them
			self.partyMemberName = nil
			self.watchedUnitID = nil
			self.threatTargetName = nil
		end
		self:UpdateThreatLists(self.threatTargetName or Threat.GlobalTarget)
	end
end

function Omen:UpdateThreatLists(target)
	if not target then
		self.bars["_AGGRO_GAIN_"].threatVal = 0
	end
	target = target or Threat.GlobalTarget
	self:Clear()
	self.Anchor.title:SetText(target and string.format("%s - |cffFF4200%s", self.modName, target) or "")
	-- self.Anchor.mobName:SetText(target and string.format("|cffFF4200%s", target) or "")
	for k, v in Threat:IterateGroupThreatForTarget(target) do
		if not self.bars[k] then
			local unit = Threat:GetUnitIDByUnitName(k)
			if unit then
				addBar(self, k, unit)
				if unit == "player" then
					self.bars[k].myArrow:Show()
				end				
				local foreign = ""
				if Threat:IsUsingForeignThreatSource(k) then
					foreign = "*"
				end
				self.bars[k].labels.name:SetText(k .. foreign)
			end
		end

		if self.bars[k] then -- unit can be non-existent when self or someone leaves party
			local val, isK = ConvertToK(v)
			self.bars[k].threatVal = v
			self.bars[k].labels.threat:SetText(string.format("%2.2f%s", val, isK))
		end
	end

	self:TriggerEvent("Omen_ThreatUpdated")
end

function Omen:Clear()
	if not self.bars then return end
	for k, v in pairs(self.bars) do
		v.threatVal = 0
		v:Hide()
	end
	self.Anchor:SetHeight(self.barsOffset + 5)
end

function Omen:PLAYER_TARGET_CHANGED()
	self.threatTargetName = UnitName("target")
	if self.threatTargetName then
		self.watchedUnitID = "target"
	else
		self.watchedUnitID = nil
	end
	self.lastWithAggro = UnitName("targettarget")
	self:UpdateThreatLists(self.threatTargetName)
end

function Omen:UNIT_TARGET(arg1)
	if UnitIsUnit("player", arg1) then
		return
		
	-- Anyone know the conditions for getting a unit named "npc"?
	elseif self.watchedUnitID and UnitIsUnit(arg1, self.watchedUnitID) and arg1 ~= "npc" then
		local lastThreatTarget = self.threatTargetName
		if UnitInParty(arg1) or UnitInRaid(arg1) or UnitIsUnit(arg1, "pet") then
			local et = arg1 .. "target"
			local ett = arg1 .. "targettarget"
			self.threatTargetName = UnitName(et)
			if UnitInParty(ett) or UnitInRaid(ett) or UnitIsUnit(ett, "pet") then
				self.lastWithAggro = UnitName(ett)
			end
			self:UpdateThreatLists(self.threatTargetName)
		else
			self.threatTargetName = UnitName(arg1)
			local et = arg1 .. "target"
			if UnitInParty(et) or UnitInRaid(et) or UnitIsUnit(et, "pet") then
				self.lastWithAggro = UnitName(arg1 .. "target")
			end
		end
		if lastThreatTarget ~= self.threatTargetName then
			self:UpdateThreatLists(self.threatTargetName)
		end
	end	
end

function Omen:Threat_ThreatUpdated(pName, pUnit, target, threatVal)
	-- self:Print(string.format("Got threat update for %s (%s): %s, %s", pName, pUnit, target, threatVal))
	if not self:IsEventScheduled("OmenTPSUpdate") then
		self:EnableTPSUpdate()
	end
	self:ScheduleEvent("ResetTPSTimerTables", resetTPSSigmas, 2, self)
	
	if not target then
		self:Print(string.format("Error: Got a nil target, %s, %s, %s", pName, pUnit, threatVal))
		return
	end
	local isCurrentTarget = self.threatTargetName and Threat:UnitIsUnit(self.threatTargetName, target)
	if not isCurrentTarget and target ~= Threat.GlobalTarget then return end

	if not self.bars[pName] then
		addBar(self, pName, pUnit)
		if pUnit == "player" then
			self.bars[pName].myArrow:Show()
		end
		local foreign = ""
		if Threat:IsUsingForeignThreatSource(pName) then
			foreign = "*"
		end
		self.bars[pName].isForeign = Threat:IsUsingForeignThreatSource(pName)
		self.bars[pName].labels.name:SetText(pName .. foreign)
	end
	local isK = ""
	if self.bars[pName].threatVal ~= threatVal then
		self.bars[pName].threatVal = tonumber(threatVal)
		threatVal, isK = ConvertToK(threatVal)
		self.bars[pName].labels.threat:SetText(string.format("%2.1f%s", threatVal, isK))
		self:TriggerEvent("Omen_ThreatUpdated")
	end
end

function Omen:Threat_OutOfDateNotice(myVersion, newVersion, newVersionHolder)
	self.Anchor.outofDate:SetText(string.format("|cffff0000New version available! |cffFFB400r |cffffffff%s (%s)", newVersion, newVersionHolder))
	self.Anchor.outofDate:Show()
end

function Omen:CreateAggroGainBar(pName, name, multiplier)
	if not self.bars[pName] then
		self.bars[pName] = self:CreateBar()
		Omen.db.profile.CustomColors[pName] = Omen.db.profile.CustomColors[pName] or {r = 1, g = 0, b = 0}
		self.bars[pName].labels.name:SetText(name)
		self.bars[pName].labels.tps:SetText("")
		self.bars[pName].unitname = pName
		tinsert(self.barList,self.bars[pName])
		self:UpdateBarDisplays()
	end
end

function Omen:UpdateAggroGainBar(pName, multiplier)
	local aggroToBeat = 0
	
	if not self.bars[pName] then return end

	if not self.watchedUnitID then return end
	for k, v in pairs(self.barList) do
		v.aggroArrow:Hide()
		if v.unitid and v.unitname == self.lastWithAggro then
			aggroToBeat = v.threatVal
			v.aggroArrow:Show()
			if Threat:UnitInMeleeRange(self.watchedUnitID) then
				multiplier = 1.1
			end
		end
	end	

	self.bars[pName].threatVal = aggroToBeat * multiplier
	local threatVal, isK = ConvertToK(self.bars[pName].threatVal)
	self.bars[pName].labels.threat:SetText(string.format("%2.1f%s", threatVal, isK))
	return aggroToBeat
end

function Omen:ShowTest()
	local pNum = GetNumPartyMembers()
	local rNum = GetNumRaidMembers()

	local uName = UnitName("player")
	if not self.bars[pName] then
		addBar(self, uName, "player")
		self.bars[uName].labels.name:SetText(uName)
	end

	uName = UnitName("pet")
	if uName then
		addBar(self, uName, "pet")
		self.bars[uName].labels.name:SetText(uName)
	end

	for i = 1, pNum do
		local pName = UnitName("party" .. i)
		if pName and not self.bars[pName] then
			addBar(self, pName, "party" .. i)
			self.bars[pName].labels.name:SetText(pName)
		end
		local pName = UnitName("party" .. i .. "pet")
		if pName and not self.bars[pName] then
			addBar(self, pName, "party" .. i .. "pet")
			self.bars[pName].labels.name:SetText(pName)
		end
	end
	
	for i = 1, rNum do
		local pName = UnitName("raid" .. i)
		if pName and not self.bars[pName] then
			addBar(self, pName, "raid" .. i)
			self.bars[pName].labels.name:SetText(pName)
		end
		local pName = UnitName("raid" .. i .. "pet")
		if pName and not self.bars[pName] then
			addBar(self, pName, "raid" .. i .. "pet")
			self.bars[pName].labels.name:SetText(pName)
		end
	end	
	
	for i = 1, (self.db.profile.MaxBars - #self.barList) do
		local pName = "* Test-" .. i
		addBar(self, pName, "player")
		self.bars[pName].labels.name:SetText(pName)
	end
		
	table.sort(self.barList, function(a,b)
		return a.unitname < b.unitname
	end)

	local lastVal = 75
	for _, v in ipairs(self.barList) do
		lastVal = lastVal + 25
		v.threatVal = lastVal
		v.tex:SetWidth(0)
		v.labels.threat:SetText(v.threatVal)
	end
	local v = self.bars["_AGGRO_GAIN_"]
	v.threatVal = lastVal + 25
	v.tex:SetWidth(0)
	v.labels.threat:SetText(v.threatVal)
	
	self:ArrangeBars()
end

function Omen:ArrangeBars()	
	if not self.Anchor:IsVisible() then return end
	local threatTarget = self.watchedUnitID

	local aggroHolderAggro = 0
	if threatTarget then	
		if Threat:UnitInMeleeRange(threatTarget) then
			aggroHolderAggro = self:UpdateAggroGainBar("_AGGRO_GAIN_", 1.1)
		else
			aggroHolderAggro = self:UpdateAggroGainBar("_AGGRO_GAIN_", 1.3)
		end
	--[[elseif self.bars["_AGGRO_GAIN_"] and not self.testing then
		self.bars["_AGGRO_GAIN_"].threatVal = 0
		]]--
	end
	local growUp = self.db.profile.GrowUp
	table.sort(self.barList, function(a,b)
		if growUp then
			return a.threatVal < b.threatVal
		else
			return b.threatVal < a.threatVal
		end
	end)
	local lastBar = nil
	local maxThreat = -1
	local barCt = 0
	local tHt = 0
	local gotPlayer = false
	local showPlayer = Omen.db.profile.AlwaysShowSelf
	local maxBars = self.db.profile.MaxBars
	local showKTM = self.db.profile.ShowKtmData
	
	if showPlayer then maxBars = maxBars - 1 end
	if #self.barList > 0 then	
		if growUp then
			maxThreat = self.barList[#self.barList].threatVal	
		else
			maxThreat = self.barList[1].threatVal
		end
	end
	
	for k,v in pairs(self.barList) do
		local isPlayer = v.unitid == "player" and showPlayer and v.threatVal > 0
		if isPlayer or (barCt < maxBars and v.threatVal > 0 and (showKTM or not v.isForeign)) then
			v:Show()
			v:ClearAllPoints()
			v:SetPercent(v.threatVal / maxThreat, v.threatVal / aggroHolderAggro)
			-- self:Print("Bar is " .. tostring(v.unitname))
			if growUp then
				if lastBar then
					v:SetPoint("BOTTOMLEFT", lastBar, "TOPLEFT", 0, 0)
					v:SetPoint("BOTTOMRIGHT", lastBar, "TOPRIGHT", 0, 0)
				else
					v:SetPoint("BOTTOMLEFT", self.Anchor, "BOTTOMLEFT", 5, 5)
					v:SetPoint("BOTTOMRIGHT", self.Anchor, "BOTTOMRIGHT", -5, 5)
				end
			else
				if lastBar then
					v:SetPoint("TOPLEFT", lastBar, "BOTTOMLEFT", 0, 0)
					v:SetPoint("TOPRIGHT", lastBar, "BOTTOMRIGHT", 0, 0)
				else
					v:SetPoint("TOPLEFT", self.Anchor, "TOPLEFT", 5, self.barsOffset * -1)
					v:SetPoint("TOPRIGHT", self.Anchor, "TOPRIGHT", -5, self.barsOffset * -1)
				end
			end
			lastBar = v
			tHt = tHt + v:GetHeight()
			if isPlayer then
				maxBars = maxBars + 1
			end
			barCt = barCt + 1
		else
			v:Hide()
		end
	end
	-- self.db.profile.BarHeight * max(0,barCt)
	self.Anchor:SetHeight(tHt + self.barsOffset + 5)
end

function Omen:OnClick()
	self:Toggle()
end

------------------------------------------------
-- UI Manipulation
------------------------------------------------

local function animateBar()
	this.animatedTime = this.animatedTime + arg1
	local pct = min(1, this.animatedTime / this.totalAnimateTime)
	local delta = (this.destinationWidth - this.startWidth) * pct
	this.tex:SetWidth(this.startWidth + delta)
	if not Omen.db.profile.StretchBarTextures then
		local texPct = (this.startWidth + delta) / this.totalWidth
		this.tex:SetTexCoord(0, texPct, 0, 1)
	end
	if pct == 1 then
		this:SetScript("OnUpdate", nil)
	end
end

local function animatePercent(self, pct, dPct) 
	local w = self:GetParent():GetWidth() - 10
	self.labels.threatpercent:SetText(string.format("%2.1f%%", dPct * 100))
	local width = w * pct
	if Omen.db.profile.AnimateBars then
		self:SetScript("OnUpdate", animateBar)
		self.totalWidth = w
		self.startWidth = self.tex:GetWidth()
		self.destinationWidth = width
		self.totalAnimateTime = 0.25
		self.animatedTime = 0
	else
		self.tex:SetWidth(width)
		if not Omen.db.profile.StretchBarTextures then
			self.tex:SetTexCoord(0, pct, 0, 1)
		end
	end	
end

function Omen:UpdateBarDisplays()
	for k, v in pairs(self.bars) do
		self:UpdateBarDisplay(k)
	end
end

local default_color = {r=0,g=1,b=1,a=1}
function Omen:UpdateBarDisplay(name)
	local v = self.bars[name]
	if not v then return end

	local texture = SML:Fetch("statusbar", Omen.db.profile.CustomBarTextures[name] or Omen.db.profile.BarTexture)
	v.tex:SetTexture(texture)
	
	local c
	if v.unitid then
		c = Omen.db.profile.CustomColors[name] or RAID_CLASS_COLORS[select(2, UnitClass(v.unitid))] or default_color
	else
		c = Omen.db.profile.CustomColors[name] or default_color
	end
	v.tex:SetVertexColor(c.r, c.g, c.b)
	
	local h = Omen.db.profile.CustomBarHeights[name] or Omen.db.profile.BarHeight
	v:SetHeight(h)
	self:ArrangeBars()
end

function Omen:CreateBar()
	local f = CreateFrame("Button", "OmenBar" .. #self.bars, self.Anchor)
	f:SetHeight(self.db.profile.BarHeight)
	
	f:SetHighlightTexture("Interface\\Buttons\\UI-Listbox-Highlight")
	-- f:GetHighlightTexture():SetBlendMode("ADD")
	
	f:RegisterForDrag("LeftButton")
	f:RegisterForClicks("RightButtonDown")
	f:SetScript("OnClick", function()
		if arg1 == "RightButton" then
			Omen.lastBarClicked = this
			dewdrop:Open(this, "children", Omen.barMenu, 'cursorX', true, 'cursorY', true)
		end
	end)
	
	f:SetScript("OnDragStart", function()
		if not this.unitid or this.unitid == "player" then return end
		local bar = OmenThreatBar:new()
		bar:CreateBar(this.unitname, this.unitid, Omen.db.profile.BarTexture)
		bar:Appear()

		bar.Bar:ClearAllPoints()
		local x, y = GetCursorPosition()
		x = x / UIParent:GetScale() - 10
		y = y / UIParent:GetScale() - 10
		bar.Bar:SetPoint("CENTER", UIParent, "BOTTOMLEFT", x, y)
		bar.Bar:SetFrameStrata("HIGH")
		bar.Bar:StartMoving()
		bar.Bar:SetFrameLevel(5)
		
		bar.Bar:SetScript("OnReceiveDrag", function()
			this:StopMovingOrSizing()
		end)

		local level = bar.Bar:GetFrameLevel()
	end)
	
	f.tex = f:CreateTexture()
	f.tex:SetPoint("TOPLEFT", f, "TOPLEFT", 0, 0)
	f.tex:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT", 0, 0)
	f.tex:SetWidth(25)
	
	f.myArrow = f:CreateTexture()
	f.myArrow:SetTexture("Interface\\AddOns\\Omen\\arrow")
	f.myArrow:SetPoint("RIGHT", f, "LEFT", -5, 0)
	f.myArrow:SetTexCoord(0, 0.5, 0, 1)
	f.myArrow:SetWidth(6)
	f.myArrow:SetHeight(12)
	f.myArrow:Hide()
	
	f.aggroArrow = f:CreateTexture()
	f.aggroArrow:SetTexture("Interface\\AddOns\\Omen\\arrow")
	f.aggroArrow:SetPoint("LEFT", f, "RIGHT", 5, 0)
	f.aggroArrow:SetTexCoord(0.5, 1, 0, 1)
	f.aggroArrow:SetWidth(6)
	f.aggroArrow:SetHeight(12)	
	f.aggroArrow:Hide()
	
	f.threatVal = 0
	
	self:CreateLabels(f)

	f.SetPercent = animatePercent
	return f
end

local function getOffset(self, data)
	local bw = self.Anchor:GetWidth() - 8
	local amt = data[1] * bw
	if data[3] == "RIGHT" then
		amt = amt * -1
	end
	return amt
end

function Omen:UpdateLayout()
	for name, column in pairs(self.Anchor.columns) do
		local posData = self.db.profile.ColumnOffsets[name]
		column:ClearAllPoints()
		column:SetPoint("TOP" .. posData[2], self.Anchor, "TOP" .. posData[3], getOffset(self, posData), self.titleOffset * -1 - 7)	

		for k, bar in pairs(self.bars) do
			local label = bar.labels[name]
			label:ClearAllPoints()
			label:SetPoint(posData[2], bar, posData[3], getOffset(self, posData), 0)	
		end
	end
	self:ArrangeBars()
end

function Omen:SetAnchorWidth(width)
	self.Anchor:SetWidth(width)
	self:UpdateLayout()
end

function Omen:CreateColumn(name, label, labelSize, dataSize)
	local t = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	t:SetVertexColor(1,1,1,1)
	local font, size = t:GetFont()
	t:SetFont(font, labelSize)
	t:SetText(label)
	
	t.name = name
	t.dataSize = dataSize
	t.offset = offset
	
	-- t.Drag = dragLabel	
	-- t.SetScript("OnMouseDown", beginLabelDragOrMenu)
	-- t.SetScript("OnMouseUp", endLabelDrag)
	
	local posData = self.db.profile.ColumnOffsets[name]
	t:SetPoint("TOP" .. posData[2], self.Anchor, "TOP" .. posData[3], getOffset(self, posData), self.titleOffset * -1 - 7)	
	self.Anchor.columns[name] = t
end

function Omen:CreateLabels(bar)
	bar.labels = {}
	local c = self.Anchor.columns
	for name, column in pairs(c) do
		local t = bar:CreateFontString(nil, nil, "GameFontNormal")
		t:SetVertexColor(1,1,1,1)
		local font, size = t:GetFont()
		t:SetFont(font, column.dataSize)
		t:SetText("-" .. name .. "-")
		
		local posData = self.db.profile.ColumnOffsets[name]
		t:SetPoint(posData[2], bar, posData[3], getOffset(self, posData), 0)
		if not self.db.profile.ColumnVisibility[name] then
			t:Hide()
		end
		bar.labels[name] = t
	end
end

function Omen:ToggleColumn(name, value)
	self.db.profile.ColumnVisibility[name] = value
	if value then
		self.Anchor.columns[name]:Show()
		for k,v in pairs(self.bars) do
			v.labels[name]:Show()
		end
	else
		self.Anchor.columns[name]:Hide()
		for k,v in pairs(self.bars) do
			v.labels[name]:Hide()
		end
	end
end

function Omen:CreateTPSGraph(unitNameA, unitNameB)
end

------------------------------------------------------------------------------------------------
-- TPS crap
------------------------------------------------------------------------------------------------

function Omen:EnableTPSUpdate()
	if not self:IsEventScheduled("OmenTPSUpdate") then
		self:ScheduleRepeatingEvent("OmenTPSUpdate", self.UpdateAllTPS, Omen.db.profile.TPSUpdateFrequency, self)
	end
end

function Omen:DisableTPSUpdate()
	if self:IsEventScheduled("OmenTPSUpdate") then
		self:CancelScheduledEvent("OmenTPSUpdate")
	end
end

function Omen:UpdateAllTPS()
	local targetName = self.threatTargetName
	if not targetName then return end
	for k, v in pairs(self.bars) do
		if v.threatVal and v.threatVal > 0 and v.unitid then
			self:UpdateTPS(k, targetName)
		end
	end	
end

function Omen:GetTPS(player, target)
	local n = tostring(target) .. "-" .. tostring(player)
	return tpsVals[n] or 0
end

-- Thanks to Kalman for working out the TPS formula with me.
-- It's just a single-pole IIR filter which decreases the weight of older samples
-- over time. 0 < p1 < 1 - a higher p1 results in faster decay (which means a shorter window)
function Omen:UpdateTPS(player, target)
	local n = target .. "-" .. player
	local sigma = tpsSigma[n] or 0
	local targetThreat = Threat:GetThreat(player, target)
	
	local freq = self.db.profile.TPSUpdateFrequency
	local beta = (2.0 - freq) / 1.95
	
	local minP1 = 0.5
	local maxP1 = 0.995
	
	-- Scale p1 linearlly from minP1 to maxP1 inverse to frequency
	local p1 = minP1 + ((maxP1 - minP1) * beta)
	if not (p1 > 0 and p1 < 1) then
		error("p1 is " .. p1 .. ", beta is " .. beta)
	end
	
	local last = (lastThreat[n] or 0)
	local delta = targetThreat - last
	lastThreat[n] = targetThreat
	if targetThreat + targetThreat < last  then
		tpsSigma[n] = 0
	else
		tpsSigma[n] = (sigma * p1) + delta
	end
	local tpsDisplay = tpsSigma[n] * (1 - p1) * (1 / freq)
	tpsVals[n] = tpsDisplay
	self.bars[player].labels.tps:SetText(string.format("%2.0f", tpsDisplay))
end

------------------------------------------------------------------------------------
-- Graphing
------------------------------------------------------------------------------------
function Omen:UpdateGraphCompare(unit)
	self.graph_unitA = unit

	self.graph_unitB = PlayerName
	if not self.GraphFrame then
		self:CreateGraph()
	end

	local r, g, b = self.bars[self.graph_unitA].tex:GetVertexColor()
	self.GraphA:SetBarColors({r, g, b, 0.1},{r, g, b, 0.6})
end

function Omen:CreateGraph()
	self.GraphFrame = CreateFrame("Frame", "OmenThreatGraph", self.Anchor)
	self.GraphFrame:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = true,
		tileSize = 16,
		edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4}
	})

	local f = self.db.profile.BackdropColor
	self.GraphFrame:SetBackdropColor(f.r, f.g, f.b, f.a)

	local f = self.db.profile.BorderColor
	self.GraphFrame:SetBackdropBorderColor(f.r, f.g, f.b, f.a)
	
	self.GraphFrame:SetWidth(self.Anchor:GetWidth())
	self.GraphFrame:SetHeight(50)
	self.GraphFrame:SetPoint("TOPLEFT", self.Anchor, "BOTTOMLEFT", 0, -10)
	local width = math.floor(self.GraphFrame:GetWidth() - 7)
	local height = math.floor(self.GraphFrame:GetHeight() - 7)
	
	local Graph = AceLibrary("Graph-1.0")

	self.GraphA = Graph:CreateGraphRealtime("OmenThreatGraphA", self.GraphFrame, "TOPLEFT", "TOPLEFT", 5, -4, width, height)
	self.GraphB = Graph:CreateGraphRealtime("OmenThreatGraphB", self.GraphFrame, "TOPLEFT", "TOPLEFT", 5, -4, width, height)

	self.graphMax = 0
		
	local g = self.GraphA
	g:SetAutoScale(false)
	g:SetGridSpacing(1.0,10.0)
	g:SetXAxis(-10,-1)
	g:SetFilterRadius(1.5)
	g:SetYMax(0.01)
	local cr, cg, cb = self.bars[self.graph_unitA].tex:GetVertexColor()
	g:SetBarColors({cr, cg, cb, 0.1},{cr, cg, cb, 0.6})

	g = self.GraphB
	g:SetAutoScale(false)
	g:SetGridSpacing(1.0,10.0)
	g:SetXAxis(-10,-1)
	g:SetYMax(0.01)
	g:SetFilterRadius(1.5)
	if self.bars[self.graph_unitB] then
		local cr, cg, cb = self.bars[self.graph_unitB].tex:GetVertexColor()
		g:SetBarColors({cr, cg, cb, 0.1},{cr, cg, cb, 0.6})
	else
		local cr, cg, cb = 1, 0, 0
		g:SetBarColors({cr, cg, cb, 0.1},{cr, cg, cb, 0.6})
	end
	self.updateCt = 9
	self.tA = 0
	self.tB = 0	
	self.tSA = 0
	self.tSB = 0	
	
	self:ScheduleRepeatingEvent("OmenUpdateThreatGraphA", function()
		self.updateCt = self.updateCt + 1
		if self.updateCt == 10 then
			local tpsA = self:GetTPS(self.graph_unitA, self.threatTargetName or "") or 0
			local tpsB = self:GetTPS(self.graph_unitB, self.threatTargetName or "") or 0
			local yMax = max(self.GraphA:GetMaxValue(), self.GraphB:GetMaxValue()) * 1.1
			self.oldTPSA = tpsA
			self.oldTPSB = tpsB
			self.GraphA.YMax = yMax
			self.GraphB.YMax = yMax
			self.tSA = (tpsA - self.tA) / 10
			self.tSB = (tpsB - self.tB) / 10
			self.tA = tpsA
			self.tB = tpsB
			self.updateCt = 0
		end
		self.GraphA:AddTimeData(self.tA + (self.tSA * self.updateCt))
		self.GraphB:AddTimeData(self.tB + (self.tSB * self.updateCt))
	end, 0.1, self)
end

function Omen:SetAnchors()
	local t = self.db.profile.GrowUp
	if t then
		local x, y = self.Anchor:GetLeft(), self.Anchor:GetBottom()
		Omen.Anchor:ClearAllPoints()
		Omen.Anchor:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x, y)
	else
		local x, y = self.Anchor:GetLeft(), self.Anchor:GetTop()
		Omen.Anchor:ClearAllPoints()
		Omen.Anchor:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x, y)
	end
end
